"""
parameter message type.
"""
import dataclasses as dc
import typing as tp

from . import fa
from .. import ba, msg
from ..msg import td


@dc.dataclass
class ContractNewMod(ba.Mod):

    def get_margin(self, value: float) -> float:
        raise NotImplementedError

    def get_value(self, value: float) -> float:
        raise NotImplementedError

    def get_order_num(self, value: float, price: float) -> float:
        raise NotImplementedError

    pass


@dc.dataclass
class Contract(ContractNewMod):
    num_per_unit: float
    value_per_dot: float
    margin_ratio: float

    def get_margin(self, value: float) -> float:
        return abs(self.get_value(value) * self.margin_ratio)

    def get_value(self, value: float):
        return value * self.value_per_dot

    def get_order_num(self, value: float, price: float) -> float:
        return (value / self.margin_ratio / self.value_per_dot / price // self.num_per_unit) * self.num_per_unit


@dc.dataclass
class ContractNewMsg(fa.S):
    """
    be decided by the exchange.
    """
    cm: ContractNewMod


class ContractDelMsg(fa.S):
    pass


@dc.dataclass
class Cash(ba.Msg):
    """
    be decided by broker.
    """
    num: float
    pass


@dc.dataclass
class CommissionMod(ba.Mod):
    def get(self, c: ContractNewMod, ts: tp.Deque[td.Trade]):
        raise NotImplementedError

    pass


@dc.dataclass
class CommissionFuture(CommissionMod):
    rate: float

    def get(self, c: ContractNewMsg, ts: tp.Deque[td.Trade]) -> float:
        return sum((t.num for t in ts)) * self.rate

    pass


@dc.dataclass
class CommissionStockA(CommissionMod):
    tax: float
    commission: float
    min_commission: float

    def get(self, c: ContractNewMod, ts: tp.Deque[td.Trade]) -> float:
        return sum([(c.get_value(abs(t.num) * t.price * self.tax) if t.num < 0 else 0) +
                    max(self.min_commission, c.get_value(abs(t.num) * t.price) * self.commission)
                    for t in ts])

    pass


@dc.dataclass
class CommissionMsg(fa.S):
    """
    be decided by broker.
    """
    cm: CommissionMod
    pass


@dc.dataclass
class SlippageMod(ba.Mod):
    """
    be decided by exchange.
    """

    def get(self, c: ContractNewMod, o: td.OrderData, price):
        raise NotImplementedError


@dc.dataclass
class SlippageFix(SlippageMod):
    step: float

    def get(self, c: ContractNewMod, o: td.OrderData, price) -> float:
        return price + (self.step if o.num > 0 else -self.step)

    pass


@dc.dataclass
class SlippagePer(SlippageMod):
    rate: float

    def get(self, c: ContractNewMod, o: td.OrderData, price) -> float:
        return price * (1 + self.rate * (1 if o.num > 0 else -1))

    pass


@dc.dataclass
class SlippageMsg(fa.S):
    sm: SlippageMod



from . import ba, cb
from .service import account
from .msg import pa, md, td

"""
interface
"""


class IOrderRsp(ba.InterFace):
    def on_trade(self, x: td.Trade):
        raise NotImplementedError

    def on_ordered(self, x: td.Ordered):
        raise NotImplementedError

    def on_canceled(self, x: td.Canceled):
        raise NotImplementedError

    def on_rejected(self, x: td.Rejected):
        raise NotImplementedError


class IOrderReq(ba.InterFace):
    def on_order(self, x: td.OrderReq):
        raise NotImplementedError

    def on_cancel(self, x: td.Cancel):
        raise NotImplementedError


class IPaReq(ba.InterFace):
    def on_commission(self, x: pa.CommissionMsg):
        raise NotImplementedError

    def on_contract_new(self, x: pa.ContractNewMsg):
        raise NotImplementedError

    def on_contract_del(self, x: pa.ContractDelMsg):
        raise NotImplementedError

    def on_cash(self, x: pa.Cash):
        raise NotImplementedError


class IMDRtn(ba.InterFace):
    def on_tick(self, x: md.Tick):
        raise NotImplementedError

    def on_order_book(self, x: md.OrderBook):
        raise NotImplementedError

    def on_open(self, x: md.Open):
        raise NotImplementedError

    def on_close(self, x: md.Close):
        raise NotImplementedError


class IFactor(ba.InterFace):
    def on_factor(self, x: md.Factor):
        raise NotImplementedError


class IGateWayInit(ba.InterFace):
    def init(self, a: account.Account, center: cb.CallBackManager, broker: cb.CallBackManager):
        raise NotImplementedError



import enum

"""
constant data。
"""


class OC(enum.Enum):
    O = 1
    C = -1


class LS(enum.Enum):
    """
    the instrument be made of the information symbol and that long or short.
    """
    L = 1
    S = -1


class BS(enum.Enum):
    B = 1
    S = -1


if __name__ == '__main__':
    pass



"""
the actor model.
"""
import typing as tp
from collections import deque, defaultdict

from .. import ba, cb, itf, cn
from ..msg import md, td, pa
from ..service import account


class GateWay(ba.Actor, itf.IMDRtn, itf.IOrderRsp, itf.IFactor, itf.IGateWayInit):
    """
    the local gateway father class.
    """
    a: account.Account
    broker: cb.CallBackManager
    center: cb.CallBackManager

    def init(self, a: account.Account, center: cb.CallBackManager, broker: cb.CallBackManager):
        center.bind(md.Factor.key, self.on_factor)
        broker.bind(td.Trade.key, self.on_trade)
        broker.bind(td.Ordered.key, self.on_ordered)
        broker.bind(td.Canceled.key, self.on_canceled)
        broker.bind(td.Rejected.key, self.on_rejected)
        broker.bind(md.Open.key, self.on_open)
        broker.bind(md.Close.key, self.on_close)
        broker.bind(md.Tick.key, self.on_tick)
        broker.bind(md.OrderBook.key, self.on_order_book)
        self.a = a
        self.center = center
        self.broker = broker
        pass

    def on_trade(self, t: td.Trade):
        pass

    def on_ordered(self, x: td.Ordered):
        pass

    def on_canceled(self, x: td.Canceled):
        pass

    def on_rejected(self, x: td.Rejected):
        pass

    def on_open(self, x: md.Open):
        pass

    def on_close(self, x: md.Close):
        pass

    def on_tick(self, x: md.Tick):
        pass

    def on_order_book(self, x: md.OrderBook):
        pass

    def on_factor(self, x: md.Factor):
        pass

    pass


class Broker(ba.Actor, itf.IOrderReq, itf.IOrderRsp, itf.IPaReq, itf.IMDRtn):
    """
    the broker actor that account service.
    """
    a: account.Account
    center: cb.CallBackManager
    gateway: cb.CallBackManager
    exchange: cb.CallBackManager

    def __init__(self, a: account.Account,
                 center: cb.CallBackManager,
                 gateway: cb.CallBackManager,
                 exchange: cb.CallBackManager):
        center.bind(pa.ContractDelMsg.key, self.on_contract_del)
        center.bind(pa.ContractNewMsg.key, self.on_contract_new)
        center.bind(pa.CommissionMsg.key, self.on_commission)
        center.bind(pa.Cash.key, self.on_cash)
        gateway.bind(td.OrderReq.key, self.on_order)
        gateway.bind(td.Cancel.key, self.on_cancel)
        exchange.bind(td.Ordered.key, self.on_ordered)
        exchange.bind(td.Trade.key, self.on_trade)
        exchange.bind(td.Canceled.key, self.on_canceled)
        exchange.bind(td.Rejected.key, self.on_rejected)
        exchange.bind(md.Open.key, self.on_open)
        exchange.bind(md.Close.key, self.on_close)
        exchange.bind(md.Tick.key, self.on_tick)
        exchange.bind(md.OrderBook.key, self.on_order_book)
        self.a = a
        self.center = center
        self.gateway = gateway
        self.exchange = exchange

    def on_cash(self, x: pa.Cash):
        self.a.cash += x.num

    def on_commission(self, x: pa.CommissionMsg):
        self.a.commissions[x.s] = x.cm

    def on_contract_new(self, x: pa.ContractNewMsg):
        self.a.contracts[x.s] = x.cm

    def on_contract_del(self, x: pa.ContractDelMsg):
        del self.a.contracts[x.s]

    def on_order(self, o: td.OrderReq):
        _od = o.od
        _cn = self.a.contracts[_od.s]
        _cm = self.a.commissions[_od.s]
        _x = True

        if (
            _od.num == 0 or  # zero number not be allowed.
            (_od.oc == cn.OC.C and abs(self.a.get_free_position(s=_od.s, ls=_od.ls)) < abs(_od.num))
        ):
            _x = False
        if _x:
            self.a.orders[o.id] = account.Order(orq=o)
            self.gateway.route(x=td.Ordered(dt=o.dt, orq=o))
            self.exchange.route(x=o)
        else:
            self.gateway.route(x=td.Rejected(dt=o.dt, orq=o))
        pass

    def on_cancel(self, x: td.Cancel):
        self.exchange.route(x=x)
        pass

    def on_ordered(self, x: td.Ordered):
        self.gateway.route(x=x)

    def on_canceled(self, x: td.Canceled):
        self.a.orders[x.orq.id].cancel_nm = self.a.orders[x.orq.id].ing
        self._on_order_rsp(x.orq)
        self.gateway.route(x=x)

    def on_rejected(self, x: td.Rejected):
        self.a.orders[x.orq.id].reject_nm = self.a.orders[x.orq.id].ing
        self._on_order_rsp(x.orq)
        self.gateway.route(x=x)
        pass

    def on_trade(self, x: td.Trade):
        acc = self.a
        if x.orq.od.oc == cn.OC.O:
            acc.positions[x.ls][x.orq.od.s].open(x=x)
        else:
            _p = acc.positions[x.ls][x.orq.od.s]
            acc.cash += _p.close(x=x, y=acc.contracts[x.orq.od.s])
            if len(_p) == 0:
                del acc.positions[x.ls][x.orq.od.s]
        acc.orders[x.orq.id].append(x)
        self._on_order_rsp(x.orq)
        self.gateway.route(x=x)
        pass

    def _on_order_rsp(self, x: td.OrderReq):
        acc = self.a
        if acc.orders[x.id].ok:
            acc.cash -= acc.commissions[x.od.s].get(c=acc.contracts[x.od.s], ts=acc.orders[x.id])

    def on_open(self, x: md.Open):
        self.gateway.route(x=x)
        pass

    def on_close(self, x: md.Close):
        self.gateway.route(x=x)
        self.a.orders = {k: v for k, v in self.a.orders.items() if not v.ok}
        pass

    def on_tick(self, x: md.Tick):
        for i in self.a.positions.values():
            if x.s in i:
                i[x.s].price = x.price
        self.gateway.route(x)
        pass

    def on_order_book(self, x: md.OrderBook):
        self.gateway.route(x=x)
        pass

    pass


class Exchange(ba.Actor, itf.IOrderReq, itf.IMDRtn):
    """
    the exchange actor.
    """
    center: cb.CallBackManager
    broker: cb.CallBackManager
    orders: tp.DefaultDict[str, tp.DefaultDict[cn.BS, tp.Deque[td.OrderReq]]]
    ticks: tp.Dict[str, md.Tick]

    def __init__(self, center: cb.CallBackManager, broker: cb.CallBackManager):
        center.bind(md.Tick.key, self.on_tick)
        center.bind(md.OrderBook.key, self.on_order_book)
        center.bind(md.Open.key, self.on_open)
        center.bind(md.Close.key, self.on_close)
        broker.bind(td.OrderReq.key, self.on_order)
        broker.bind(td.Cancel.key, self.on_cancel)
        self.center = center
        self.broker = broker
        self.orders = defaultdict(lambda: defaultdict(deque))
        self.ticks = dict()

    def _order_match(self, o: td.OrderReq, p: float):
        _t = type(o.od)
        if _t == td.Limit:
            if (
                (o.od.price >= p and o.od.bs == cn.BS.B) or
                (o.od.price <= p and o.od.bs == cn.BS.S)
            ):
                self.broker.route(x=td.Trade(dt=o.dt, orq=o, price=p, num=o.od.num, ))
            else:
                self.orders[o.od.s][o.od.bs].append(o)
        elif _t == td.Market:
            self.broker.route(x=td.Trade(dt=o.dt, orq=o, price=p, num=o.od.num, ))

    def on_order(self, x: td.OrderReq):
        self.broker.route(td.Ordered(dt=x.dt, orq=x))
        if x.od.s in self.ticks:
            self._order_match(o=x, p=self.ticks[x.od.s].price)
        else:
            self.orders[x.od.s][x.od.bs].append(x)
        pass

    def on_cancel(self, x: td.Cancel):
        pass

    def on_open(self, x: md.Open):
        self.broker.route(x)
        pass

    def on_close(self, x: md.Close):
        self.broker.route(x)
        pass

    def on_tick(self, x: md.Tick):
        self.ticks[x.s] = x
        if x.s in self.orders:
            for bs, v in self.orders[x.s].items():
                self.orders[x.s][bs] = deque()
                for o in v:
                    self._order_match(o=o, p=x.price)
        self.broker.route(x)
        pass

    def on_order_book(self, x: md.OrderBook):
        self.broker.route(x)
        pass

    pass



"""
trade message type.
"""
import dataclasses as dc

from .. import ba, cn


class ILS(ba.InterFace):
    @property
    def ls(self) -> cn.LS:
        """

        :return: cn.LS:operate the position type：long or short
        """

        raise NotImplementedError


@dc.dataclass
class OrderData(ba.Mod, ILS):
    s: str
    oc: cn.OC
    price: float
    num: float

    @property
    def ls(self) -> cn.LS:
        return (
            cn.LS.L if (self.num > 0 and self.oc == cn.OC.O) or (self.num < 0 and self.oc == cn.OC.C)
            else cn.LS.S)

    @property
    def bs(self) -> cn.BS:
        return cn.BS.B if self.num > 0 else cn.BS.S


@dc.dataclass
class Limit(OrderData):
    pass


@dc.dataclass
class Market(OrderData):
    """
    see：https://blog.csdn.net/u012724887/article/details/98502040
    """
    pass


@dc.dataclass
class OrderReq(ba.Msg):
    od: OrderData
    pass


@dc.dataclass
class OrderRsp(ba.Msg):
    orq: OrderReq


@dc.dataclass
class Ordered(OrderRsp):
    pass


@dc.dataclass
class Cancel(ba.Msg):
    ord: Ordered
    pass


@dc.dataclass
class Rejected(OrderRsp):
    pass


@dc.dataclass
class Canceled(OrderRsp):
    pass


@dc.dataclass
class Trade(OrderRsp, ILS):
    price: float
    num: float

    @property
    def ls(self) -> cn.LS:
        return cn.LS.L if (self.num > 0 and self.orq.od.oc == cn.OC.O) or \
                          (self.num < 0 and self.orq.od.oc == cn.OC.C) \
            else cn.LS.S

    pass



import dataclasses as dc
import typing as tp
from collections import defaultdict

from .. import ba, cn
from ..msg import td, pa


@dc.dataclass
class Position(ba.Service, tp.Deque[td.Trade]):
    price: float = dc.field(init=False, default=0)

    def close(self, x: td.Trade, y: pa.ContractNewMod) -> float:
        self.price = x.price
        _volume = x.num
        _profit = 0
        _t = None
        _bs = -1 if x.num < 0 else 1
        while _volume * _bs > 0:
            if abs(self[0].num) > abs(_volume):
                _t = self[0]
                _t.num += _volume
                _profit += y.get_value(value=_volume * (x.price - _t.price))
                _volume = 0
            else:
                _t = self.popleft()
                _profit += y.get_value(value=_t.num * (x.price - _t.price))
                _volume += _t.num
        return _profit

    def open(self, x: td.Trade):
        self.price = x.price
        self.append(x)


@dc.dataclass
class Order(ba.Service, tp.Deque[td.Trade]):
    orq: td.OrderReq
    cancel_nm: float = dc.field(init=False, default=0)  # 取消
    reject_nm: float = dc.field(init=False, default=0)  # 拒绝

    @property
    def ing(self) -> float:
        return self.orq.od.num - (sum([i.num for i in self]) + self.cancel_nm + self.reject_nm)

    @property
    def ok(self) -> bool:
        return self.ing == 0

    pass


@dc.dataclass
class Account(ba.Service):
    """
    the account service.
    """
    cash: float = dc.field(default=0, init=False)
    contracts: tp.Dict[str, pa.ContractNewMod] = dc.field(default_factory=dict, init=False)
    commissions: tp.Dict[str, pa.CommissionMod] = dc.field(default_factory=dict, init=False)
    orders: tp.Dict[str, Order] = dc.field(default_factory=dict, init=False)
    positions: tp.DefaultDict[cn.LS, tp.DefaultDict[str, Position]] = dc.field(
        default_factory=lambda: defaultdict(lambda: defaultdict(Position)), init=False)

    @property
    def margin(self) -> float:
        return sum([
            self.contracts[symbol].get_margin(p.price * t.num)
            for ps in self.positions.values() for symbol, p in ps.items() for t in p
        ])

    @property
    def profit(self) -> float:
        return sum([
            self.contracts[symbol].get_value((p.price - t.price) * t.num)
            for ps in self.positions.values() for symbol, p in ps.items() for t in p
        ])

    @property
    def frozen(self) -> float:
        return sum([
            self.contracts[i.orq.od.s].get_margin(i.ing * i.orq.od.price)
            for i in self.orders.values() if i.orq.od.oc == cn.OC.O
        ])

    @property
    def equity(self) -> float:
        return self.cash + self.profit

    @property
    def free(self) -> float:
        return max(self.cash - self.frozen - self.margin + self.profit, 0)

    @property
    def pv(self) -> float:
        """

        :return: the positions value.
        """
        return sum([
            self.contracts[symbol].get_value(p.price * abs(t.num))
            for ps in self.positions.values() for symbol, p in ps.items() for t in p
        ])

    @property
    def commission(self) -> float:
        return sum([
            self.commissions[i.orq.od.s].get(c=self.contracts[i.orq.od.s], ts=i)
            for i in self.orders.values() if i.ok
        ])

    def get_free_position(self, s: str, ls: cn.LS) -> float:
        return (
            sum([t.num
                 for _ls, ps in self.positions.items() if _ls == ls
                 for _s, p in ps.items() if _s == s
                 for t in p]) -
            sum([o.ing
                 for o in self.orders.values() if o.orq.od.s == s and not o.ok])
        )

    pass



import typing as t
import uuid
from collections import defaultdict

from . import ba

CallBackType = t.Callable[[], None]
"""
the callback center.
"""


class CallBackManager(object):
    """
    manage the callbacks and msg.
    """
    id: str
    _callbacks: t.Dict[int, t.List[CallBackType]]

    def __init__(self):
        self.id = str(uuid.uuid1())
        self._callbacks = defaultdict(list)
        pass

    def bind(self, key: int, callback: CallBackType):
        call_back_list = self._callbacks[key]
        if callback not in call_back_list:
            call_back_list.append(callback)
        else:
            raise ValueError('call back function only be promised register once.')

    def unbind(self, key: int, callback: CallBackType):
        call_back_list = self._callbacks[key]
        if callback in call_back_list:
            call_back_list.remove(callback)
        else:
            raise ValueError('call back function is not in the manage center.')

    def route(self, x: ba.Msg):
        if x.key in self._callbacks:
            list(map(lambda cb: cb(x), self._callbacks[x.key]))



"""
market message type.
"""
import dataclasses as dc

from . import fa
from .. import ba


@dc.dataclass
class Tick(fa.S):
    price: float
    volume: float
    pass


@dc.dataclass
class OrderBook(fa.S):
    pass


@dc.dataclass
class Open(ba.Msg):
    pass


class Close(ba.Msg):
    pass


@dc.dataclass
class Factor(fa.S):
    data: dict
    pass



from .. import ba, cb, actor
from ..service import account


class A0(ba.App):
    center: cb.CallBackManager
    a: account.Account

    def __init__(self, stg: actor.GateWay):
        a = account.Account()
        center = cb.CallBackManager()
        stg.init(a=a, center=center, broker=cb.CallBackManager())
        actor.Broker(
            a=a, center=center,
            gateway=stg.broker,
            exchange=actor.Exchange(center=center, broker=cb.CallBackManager()).broker)
        self.center = center
        self.a = a

    def route(self, x: ba.Msg):
        self.center.route(x=x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os

from setuptools import setup

with open('README.MD') as readme_file:
    readme = readme_file.read()


def get_req(file: str) -> list:
    _r = []
    with open(os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        file
    )) as f:
        _r.extend([l for l in f.read().splitlines() if not l.startswith('#')])
    return _r


setup_requirements = []

setup(
    install_requires=get_req('requirements.txt'),
)



import dataclasses as dc
from datetime import datetime

"""
the base of project class.
"""


class Actor(object):
    """
    角色
    """
    pass


class App(object):
    """
    应用：模拟交易
    """
    pass


_uid = 0


def uid():
    global _uid
    _uid += 1
    return _uid


@dc.dataclass
class Msg(object):
    """
    消息
    """
    key: int = dc.field(default=None, init=False, repr=False)
    id: str = dc.field(default_factory=lambda: str(uid()), init=False, repr=False)
    dt: datetime

    def __init_subclass__(cls, **kwargs):
        cls.key = hash(cls)


class Mod(object):
    """
    数据模型
    """
    pass


class Service(object):
    """
    服务
    """
    pass


class InterFace(object):
    """
    接口
    """
    pass
